from json import dumps, loads
from functools import wraps
from .basecache import BaseCache, BaseCacheDecorator


class NoCache(BaseCache):
    def __init__(self, prefix="rc", serializer=dumps, deserializer=loads):
        super().__init__(NoCacheDecorator, dict(), prefix, serializer, deserializer)

    def mget(self, *fns_with_args):
        raise NotImplementedError('This is not a real cache..')


class NoCacheDecorator(BaseCacheDecorator):

    def __call__(self, fn):

        @wraps(fn)
        def inner(*args, **kwargs):
            return fn(*args, **kwargs)

        inner.invalidate = self.invalidate
        inner.invalidate_all = self.invalidate_all
        inner.instance = self
        return inner

    def check_cache(self, key):
        return None

    def cache_output(self, key, serialized):
        pass

    def invalidate_key(self, key):
        pass




from json import dumps, loads
from time import time
from .basecache import BaseCache, BaseCacheDecorator


class CachedItem(object):
    def __init__(self, key, value, duration=60):
        self.key = key
        self.value = value
        self.duration = duration
        self.timestamp = time()

    def expired(self):
        if self.duration == 0:
            return False
        return self.timestamp + self.duration < time()

    def __repr__(self):
        return '<CachedItem {%s:%s} expires at: %s>' % (self.key, self.value, self.timestamp + self.duration)


class CachedDict(dict):
    def __init__(self, seq=None, **kwargs):
        super().__init__(seq=None, **kwargs)
        self.operations = 0
        self.prune_operations = 50

    def _prune_if(self):
        self.operations += 1
        if self.operations < self.prune_operations:
            return
        self.operations = 0
        obsolete = [k for k, v in self.items() if v and v.expired()]
        for k in obsolete:
            del self[k]

    def get(self, key):
        self._prune_if()
        if key not in self:
            return None
        elif self[key].expired():
            del self[key]
            return None
        else:
            return self[key].value

    def set(self, key, value, duration=60):
        self[key] = CachedItem(key, value, duration)
        self._prune_if()


class MemCache(BaseCache):
    def __init__(self, prefix="rc", serializer=dumps, deserializer=loads):
        super().__init__(MemCacheDecorator, CachedDict(), prefix, serializer, deserializer)

    def mget(self, *fns_with_args):
        keys = self.mget_keys(*fns_with_args)
        results = {key: self._cache.get(key) for key in keys}
        return [self.deserializer(v) for k,v in results.items() if v]


class MemCacheDecorator(BaseCacheDecorator):
    def __init__(self, cache, prefix='rc', serializer=dumps, deserializer=loads, ttl=0, limit=0, namespace=None):
        if limit != 0:
            raise ValueError('MemCache does not support limits - only ttl')
        super().__init__(cache, prefix, serializer, deserializer, ttl, limit, namespace)

    def check_cache(self, key):
        return self.cache.get(key)

    def cache_output(self, key, serialized):
        self.cache.set(key, serialized, self.ttl)

    def invalidate_key(self, key):
        try:
            del self.cache[key]
        except KeyError:
            pass  # already invalidated..

    def invalidate_all(self, *args, **kwargs):
        if not self.namespace or not self.cache:
            return
        invalidated = [k for k in self.cache if self.namespace in k]
        for k in invalidated:
            del self.cache[k]



from .rediscache import RedisCache
from .memcache import MemCache
from .diskcache import DiskCache
from .nocache import NoCache

DEFAULT_SETTINGS = {
    'type': 'MemCache',
    'prefix': 'rc',
    'serializer': 'json.dumps',
    'deserializer': 'json.loads',
    'diskcache_directory': None,  # diskcache will default to os specific temp dir
    'redis_host': 'localhost',
    'redis_port': 6379,
    'redis_db': 0,
    'redis_username': None,
    'redis_password': None,
    'redis_clientname': None,
    'redis_ssl_check_hostname': None,
    'redis_decode_responses': True,
}


def _load_func(text):
    from re import match
    m = match(r'^([a-z].*)\.([\w_]+)$', text)
    if not m:
        raise ValueError('\'{}\' is not a valid module.method string?'.format(text))
    module = __import__(m.group(1))
    return getattr(module, m.group(2))


def init_cache_from_settings(settings):
    """
    Initialize a cache using settings dict (see flex_cache.DEFAULT_SETTINGS)
    Args:
        settings:

    Returns:

    """
    merged = DEFAULT_SETTINGS.copy()
    merged.update(settings)
    common_kwargs = {'prefix': merged['prefix'],
                     'serializer': _load_func(merged['serializer']),
                     'deserializer': _load_func(merged['deserializer']),
                     }
    if merged['type'] == 'MemCache':
        return MemCache(**common_kwargs)
    elif merged['type'] == 'DiskCache':
        from diskcache import Cache as DCache
        dc = DCache(directory=merged['diskcache_directory'])
        return DiskCache(dc, **common_kwargs)
    elif merged['type'] == 'RedisCache':
        redis_kwargs = {k[6:]: v for k, v in common_kwargs.items() if k.startswith('redis')}
        from redis import Redis
        client = Redis(**redis_kwargs)
        return RedisCache(client, **common_kwargs)
    elif merged['type'] == 'NoCache':
        return NoCache()
    else:
        raise ValueError('Unsupported caching type: {}'.format(merged['type']))



from json import dumps, loads
from time import time
from diskcache import Cache as DCache
from .memcache import MemCache, MemCacheDecorator


class DiskCache(MemCache):
    """
    DiskCache inherits from MemCache since the underlying cache object is also dict-like.
    DiskCache must be initiated with a diskcache.Cache object to back the disk-based caching
    """
    def __init__(self, dcache=None, prefix="rc", serializer=dumps, deserializer=loads):
        dcache = dcache or DCache()
        # Use BaseCache init rather than the MemCache one..
        super(MemCache, self).__init__(DiskCacheDecorator, dcache, prefix, serializer, deserializer)


class DiskCacheDecorator(MemCacheDecorator):
    def __init__(self, cache, prefix='rc', serializer=dumps, deserializer=loads, ttl=0, limit=0, namespace=None):
        if limit != 0:
            raise ValueError('DiskCache does not support limits - only ttl')
        if ttl == 0:
            ttl = None
        super().__init__(cache, prefix, serializer, deserializer, ttl, limit, namespace)



from setuptools import setup, find_packages

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='python-flex-cache',
      version='0.1.6',
      description='Basic Redis/Disk/Memory caching for functions',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='http://github.com/steffenschumacher/python-flex-cache',
      author='Steffen Schumacher (forked from python-redis-cache // Taylor Hakes)',
      license='MIT',
      python_requires='>=3.6',
      packages=find_packages(),
      setup_requires=['pytest-runner==5.2', 'diskcache==5.2.1'],
      tests_require=['pytest==5.4.3', 'redis==3.5.3'],
)



from functools import wraps
from json import dumps, loads
from base64 import b64encode


class BaseCache:
    def __init__(self, decorator, cache, prefix="rc", serializer=dumps, deserializer=loads):
        self._decorator = decorator
        self._cache = cache
        self.prefix = prefix
        self.serializer = serializer
        self.deserializer = deserializer

    def cache(self, ttl=0, limit=0, namespace=None):
        return self._decorator(self._cache, self.prefix, self.serializer, self.deserializer, ttl, limit, namespace)

    def mget_keys(self, *fns_with_args):
        keys = []
        for fn_and_args in fns_with_args:
            fn = fn_and_args['fn']
            args = fn_and_args['args'] if 'args' in fn_and_args else []
            kwargs = fn_and_args['kwargs'] if 'kwargs' in fn_and_args else {}
            keys.append(fn.instance.get_key(args=args, kwargs=kwargs))
        return keys

    def mget(self, *fns_with_args):
        raise NotImplementedError('Must be implemented in derived classes')

    def _key(self, key, namespace=None):
        if not isinstance(key, str):
            key = str(b64encode(key), 'utf-8')
        if namespace:
            return f'{self.prefix}:{namespace}:{key}'
        else:
            return f'{self.prefix}:{key}'

    def get(self, key, namespace=None):
        serialized = self.cache(namespace=namespace).check_cache(self._key(key, namespace))
        if serialized:
            return self.deserializer(serialized)
        else:
            return None

    def set(self, key, value, ttl=0, limit=0, namespace=None):
        serialized = self.serializer(value)
        deco = self.cache(ttl, limit, namespace)
        deco.keys_key = self._key('keys', namespace=namespace)
        return deco.cache_output(self._key(key, namespace), serialized)

    def invalidate(self, key, namespace=None):
        deco = self.cache(namespace=namespace)
        deco.keys_key = self._key('keys', namespace=namespace)
        deco.invalidate_key(self._key(key, namespace))


class BaseCacheDecorator:
    def __init__(self, cache, prefix='rc', serializer=dumps, deserializer=loads, ttl=0, limit=0, namespace=None):
        self.cache = cache
        self.prefix = prefix
        self.serializer = serializer
        self.deserializer = deserializer
        self.ttl = ttl
        self.limit = limit
        self.namespace = namespace
        self.keys_key = None

    def get_key(self, args, kwargs):
        serialized_data = self.serializer([args, kwargs])

        if not isinstance(serialized_data, str):
            serialized_data = str(b64encode(serialized_data), 'utf-8')
        return f'{self.prefix}:{self.namespace}:{serialized_data}'

    def __call__(self, fn):
        self.namespace = self.namespace if self.namespace else f'{fn.__module__}.{fn.__name__}'
        self.keys_key = f'{self.prefix}:{self.namespace}:keys'
        self.original_fn = fn

        @wraps(fn)
        def inner(*args, **kwargs):
            nonlocal self
            key = self.get_key(args, kwargs)
            result = self.check_cache(key)
            if not result:
                result = fn(*args, **kwargs)
                result_serialized = self.serializer(result)
                self.cache_output(key, result_serialized)
            else:
                result = self.deserializer(result)
            return result

        inner.invalidate = self.invalidate
        inner.invalidate_all = self.invalidate_all
        inner.instance = self
        return inner

    def check_cache(self, key):
        raise NotImplementedError('Must be implemented in derived classes')

    def cache_output(self, key, serialized):
        raise NotImplementedError('Must be implemented in derived classes')

    def invalidate_key(self, key):
        raise NotImplementedError('Must be implemented in derived classes')

    def invalidate(self, *args, **kwargs):
        key = self.get_key(args, kwargs)
        self.invalidate_key(key)
        if key in self.cache:
            self.invalidate_key(key)

    def invalidate_all(self, *args, **kwargs):
        if not self.namespace or not self.cache:
            return
        key_prefix = f'{self.prefix}:{self.namespace}:'
        for k in [k for k in self.cache if k.startswith(key_prefix)]:
            self.invalidate_key(k)




from json import dumps, loads
from .basecache import BaseCache, BaseCacheDecorator


def get_cache_lua_fn(client):
    if not hasattr(client, '_lua_cache_fn'):
        client._lua_cache_fn = client.register_script("""
local ttl = tonumber(ARGV[2])
local value
if ttl > 0 then
  value = redis.call('SETEX', KEYS[1], ttl, ARGV[1])
else
  value = redis.call('SET', KEYS[1], ARGV[1])
end
local limit = tonumber(ARGV[3])
if limit > 0 then
  local time_parts = redis.call('TIME')
  local time = tonumber(time_parts[1] .. '.' .. time_parts[2])
  redis.call('ZADD', KEYS[2], time, KEYS[1])
  local count = tonumber(redis.call('ZCOUNT', KEYS[2], '-inf', '+inf'))
  local over = count - limit
  if over > 0 then
    local stale_keys_and_scores = redis.call('ZPOPMIN', KEYS[2], over)
    -- Remove the the scores and just leave the keys
    local stale_keys = {}
    for i = 1, #stale_keys_and_scores, 2 do
      stale_keys[#stale_keys+1] = stale_keys_and_scores[i]
    end
    redis.call('ZREM', KEYS[2], unpack(stale_keys))
    redis.call('DEL', unpack(stale_keys))
  end
end
return value
""")
    return client._lua_cache_fn


# Utility function to batch keys
def chunks(iterable, n):
    """Yield successive n-sized chunks from iterator."""
    _iterable = iter(iterable)
    while True:
        elements = []
        for _ in range(n):
            try:
                elements.append(next(_iterable))
            except StopIteration:
                break

        if not len(elements):
            break

        yield elements


class RedisCache(BaseCache):
    def __init__(self, redis_client, prefix="rc", serializer=dumps, deserializer=loads):
        super().__init__(RedisCacheDecorator, redis_client, prefix, serializer, deserializer)

    def mget(self, *fns_with_args):
        keys = self.mget_keys(*fns_with_args)
        results = self._cache.mget(*keys)
        pipeline = self._cache.pipeline()

        deserialized_results = []
        needs_pipeline = False
        for i, result in enumerate(results):
            if result is None:
                needs_pipeline = True

                fn_and_args = fns_with_args[i]
                fn = fn_and_args['fn']
                args = fn_and_args['args'] if 'args' in fn_and_args else []
                kwargs = fn_and_args['kwargs'] if 'kwargs' in fn_and_args else {}
                result = fn.instance.original_fn(*args, **kwargs)
                result_serialized = self.serializer(result)
                get_cache_lua_fn(self._cache)(keys=[keys[i], fn.instance.keys_key], args=[result_serialized, fn.instance.ttl, fn.instance.limit], client=pipeline)
            else:
                result = self.deserializer(result)
            deserialized_results.append(result)

        if needs_pipeline:
            pipeline.execute()
        return deserialized_results


class RedisCacheDecorator(BaseCacheDecorator):
    def __init__(self, redis_client, prefix="rc", serializer=dumps, deserializer=loads, ttl=0, limit=0, namespace=None):
        super().__init__(redis_client, prefix, serializer, deserializer, ttl, limit, namespace)

    def check_cache(self, key):
        return self.cache.get(key)

    def cache_output(self, key, serialized):
        get_cache_lua_fn(self.cache)(keys=[key, self.keys_key], args=[serialized, self.ttl, self.limit])

    def invalidate_key(self, key):
        pipe = self.cache.pipeline()
        pipe.delete(key)
        pipe.zrem(self.keys_key, key)
        pipe.execute()

    def invalidate_all(self, *args, **kwargs):
        chunks_gen = chunks(self.cache.scan_iter(f'{self.prefix}:{self.namespace}:*'), 500)
        for keys in chunks_gen:
            self.cache.delete(*keys)
